# talKZ - (Kalfa and Zack)'s talk implementation
# Copyright (C) 2002:
#     Cosimo Alfarano      <kalfa@cs.unibo.it>
#     Stefano Zacchiroli   <zack@cs.unibo.it>
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
Wrapper around 'struct' module that support somewhat 'extended' format string,
see docstring of kzstruct.pack for more information
"""

import re
import struct


debug = False


class Error(Exception):

    def __init__(self, msg):
        self.__msg = msg

    def __str__(self):
        return `self.__msg`

    def getMsg(self):
        return self.__msg


def is_padding(fmt):
    """ return True if 'fmt' is a padding format string for struct.pack, False
    otherwise """
    if re.search(re.compile('x'), fmt):
        return True
    else:
        return False

def is_aligned(fmt):
    return (calcsize(fmt) % 4 == 0)

def pack(fmt, args):
    """ Wrapper around struct.pack, handle network-host byte order conversion
    on each format string component instead of on the whole format string. For
    example it can handle format strings like "B !B I i !i", translating in
    network byte order 2nd and 5th arguments but not the other one.
    
    Warning: does not support repetition other than on padding ('x') and string
    ('s') formats, e.g.: format "B 3x I" _is_ supported, but "B 3x 2I" _isn't_
    supported

    Warning: supports only aligned format string
    """

    assert is_aligned(fmt)
    fmts = re.split(re.compile(" "), fmt)
    args.reverse()  # because I use pop to extract args
    packed_data = ""
    for f in fmts:  # pack each string and catenate them
        if debug:
            try:
                print "DEBUG, fmt = %s, fmt size = %d, next arg = %s" % \
                        (f, struct.calcsize(f), str(type(args[-1])))
            except IndexError:
                print "DEBUG, fmt = %s, fmt size = %d, next arg = none" % \
                        (f, struct.calcsize(f))
        if is_padding(f):
            packed_data += struct.pack(f)
        else:   # isn't padding
            packed_data += struct.pack(f, args.pop())

    return packed_data


def unpack(fmt, string):
    """ wrapper aroung struct.unpack that support the same format strings
    supported by kzstruct.pack """

    assert is_aligned(fmt)
    if calcsize(fmt) != len(string):
        raise Error("size mismatch")
    fmts = re.split(re.compile(" "), fmt)
    args = []
    for f in fmts:
        siz = struct.calcsize(f)
        strslice = string[:siz]
        string = string[siz:]
        if not is_padding(f):   # skip padding bytes
            args.extend(struct.unpack(f, strslice))
    return tuple(args)


def calcsize(fmt):
    """ wrapper around struct.calcsize that support the same format strings
    supported by kzstruct.pack
    
    Warning: due to lack of support for unaligned format string
    kzstruct.calcsize may return a different value than struct.calcsize.
    E.g.:
        kzstruct.calcsize("I x I") = 9
        struct.calcsize("I x I") = 12
    because the padding byte is aligned to 4 byte boundary by struct module
    Anyway the result is granted to be the same if format string represent an
    aligned structure
    """

    # warning: this is _not_ equal to struct.calcsize(fmt)
    return reduce((lambda x,y: x+struct.calcsize(y)), \
                    re.split(re.compile(" "), fmt), 0)

